import { CommandHandler, ICommandHandler } from '@nestjs/cqrs';
import { CreateAdCommand } from './create-ad.command';
import { DefaultParams } from '@modules/ad/core/ports/default-params.type';
import { Inject } from '@nestjs/common';
import { AD_REPOSITORY, PARAMS_PROVIDER } from '@modules/ad/ad.di-tokens';
import { AdRepositoryPort } from '@modules/ad/core/ports/ad.repository.port';
import { DefaultParamsProviderPort } from '@modules/ad/core/ports/default-params-provider.port';
import { Err, Ok, Result } from 'oxide.ts';
import { AggregateID } from '@libs/ddd';
import { AdAlreadyExistsError } from '@modules/ad/core/ad.errors';
import { AdEntity } from '@modules/ad/core/ad.entity';
import { ConflictException } from '@libs/exceptions';
import { Waypoint } from '../../types/waypoint';

@CommandHandler(CreateAdCommand)
export class CreateAdService implements ICommandHandler {
  private readonly defaultParams: DefaultParams;

  constructor(
    @Inject(AD_REPOSITORY)
    private readonly repository: AdRepositoryPort,
    @Inject(PARAMS_PROVIDER)
    private readonly defaultParamsProvider: DefaultParamsProviderPort,
  ) {
    this.defaultParams = defaultParamsProvider.getParams();
  }

  async execute(
    command: CreateAdCommand,
  ): Promise<Result<AggregateID, AdAlreadyExistsError>> {
    const ad = AdEntity.create(
      {
        userId: command.userId,
        driver: command.driver,
        passenger: command.passenger,
        frequency: command.frequency,
        fromDate: command.fromDate,
        toDate: command.toDate,
        schedule: command.schedule,
        marginDurations: command.marginDurations,
        seatsProposed: command.seatsProposed,
        seatsRequested: command.seatsRequested,
        strict: command.strict,
        waypoints: command.waypoints.map((waypoint: Waypoint) => ({
          position: waypoint.position,
          address: {
            name: waypoint.name,
            houseNumber: waypoint.houseNumber,
            street: waypoint.street,
            postalCode: waypoint.postalCode,
            locality: waypoint.locality,
            country: waypoint.country,
            coordinates: {
              lon: waypoint.lon,
              lat: waypoint.lat,
            },
          },
        })),
      },
      {
        driver: this.defaultParams.DRIVER,
        passenger: this.defaultParams.PASSENGER,
        marginDurations: {
          mon: this.defaultParams.MON_MARGIN,
          tue: this.defaultParams.TUE_MARGIN,
          wed: this.defaultParams.WED_MARGIN,
          thu: this.defaultParams.THU_MARGIN,
          fri: this.defaultParams.FRI_MARGIN,
          sat: this.defaultParams.SAT_MARGIN,
          sun: this.defaultParams.SUN_MARGIN,
        },
        strict: this.defaultParams.STRICT,
        seatsProposed: this.defaultParams.SEATS_PROPOSED,
        seatsRequested: this.defaultParams.SEATS_REQUESTED,
      },
    );

    try {
      await this.repository.insert(ad);
      return Ok(ad.id);
    } catch (error: any) {
      if (error instanceof ConflictException) {
        return Err(new AdAlreadyExistsError(error));
      }
      throw error;
    }
  }
}
