import { Controller, UsePipes } from '@nestjs/common';
import { CommandBus } from '@nestjs/cqrs';
import { GrpcMethod, RpcException } from '@nestjs/microservices';
import { RpcValidationPipe } from '../../../../utils/pipes/rpc.validation-pipe';
import { AdPresenter } from '../ad.presenter';
import { CreateAdRequestDTO } from './dtos/create-ad.request.dto';
import { CreateAdCommand } from '../../core/commands/create-ad/create-ad.command';
import { Result, match } from 'oxide.ts';
import { AggregateID } from '@libs/ddd';
import { AdAlreadyExistsError } from '../../core/ad.errors';
import { IdResponse } from '@libs/api/id.response.dto';

@UsePipes(
  new RpcValidationPipe({
    whitelist: false,
    forbidUnknownValues: false,
  }),
)
@Controller()
export class CreateAdGrpcController {
  constructor(private readonly commandBus: CommandBus) {}

  @GrpcMethod('AdsService', 'Create')
  async create(data: CreateAdRequestDTO): Promise<AdPresenter> {
    const result: Result<AggregateID, AdAlreadyExistsError> =
      await this.commandBus.execute(new CreateAdCommand(data));

    // Deciding what to do with a Result (similar to Rust matching)
    // if Ok we return a response with an id
    // if Error decide what to do with it depending on its type
    return match(result, {
      Ok: (id: string) => new IdResponse(id),
      Err: (error: Error) => {
        if (error instanceof AdAlreadyExistsError)
          throw new RpcException({
            code: 6,
            message: 'Ad already exists',
          });
        throw new RpcException({});
      },
    });
  }
}
