import { Test } from '@nestjs/testing';
import { PrismaService } from '../../../database/adapters/secondaries/prisma-service';
import { AdsRepository } from '../../adapters/secondaries/ads.repository';
import { DatabaseModule } from '../../../database/database.module';
import { Frequency } from '../../interface/commands/frequency.enum';
import { AdDTO } from '../../core/dtos/ad.dto';
import { Waypoint } from '../../core/entities/waypoint';

describe('Ad Repository', () => {
  let prismaService: PrismaService;
  let adsRepository: AdsRepository;

  const executeInsertCommand = async (table: string, object: any) => {
    const command = `INSERT INTO ${table} ("${Object.keys(object).join(
      '","',
    )}") VALUES (${Object.values(object).join(',')})`;

    await prismaService.$executeRawUnsafe(command);
  };
  const getSeed = (index: number, uuid: string): string => {
    return `'${uuid.slice(0, -2)}${index.toString(16).padStart(2, '0')}'`;
  };
  const baseUuid = {
    uuid: 'be459a29-7a41-4c0b-b371-abe90bfb6f00',
  };
  const baseAdress0Uuid = {
    uuid: 'bad5e786-3b15-4e51-a8fc-926fa9327ff1',
  };
  const baseAdress1Uuid = {
    uuid: '4d200eb6-7389-487f-a1ca-dbc0e40381c9',
  };
  const baseUserUuid = {
    userUuid: "'113e0000-0000-4000-a000-000000000000'",
  };
  const driverAd = {
    driver: 'true',
    passenger: 'false',
    seatsDriver: 3,
    seatsPassenger: 0,
    strict: 'false',
  };
  const passengerAd = {
    driver: 'false',
    passenger: 'true',
    seatsDriver: 0,
    seatsPassenger: 1,
    strict: 'false',
  };
  const driverAndPassengerAd = {
    driver: 'true',
    passenger: 'true',
    seatsDriver: 3,
    seatsPassenger: 1,
    strict: 'false',
  };
  const punctualAd = {
    frequency: `'PUNCTUAL'`,
    fromDate: `'2023-01-01'`,
    toDate: `'2023-01-01'`,
    monTime: 'NULL',
    tueTime: 'NULL',
    wedTime: 'NULL',
    thuTime: 'NULL',
    friTime: 'NULL',
    satTime: 'NULL',
    sunTime: `'07:00'`,
    monMargin: 900,
    tueMargin: 900,
    wedMargin: 900,
    thuMargin: 900,
    friMargin: 900,
    satMargin: 900,
    sunMargin: 900,
  };
  const recurrentAd = {
    frequency: `'RECURRENT'`,
    fromDate: `'2023-01-01'`,
    toDate: `'2023-12-31'`,
    monTime: `'07:00'`,
    tueTime: `'07:00'`,
    wedTime: `'07:00'`,
    thuTime: `'07:00'`,
    friTime: `'07:00'`,
    satTime: 'NULL',
    sunTime: 'NULL',
    monMargin: 900,
    tueMargin: 900,
    wedMargin: 900,
    thuMargin: 900,
    friMargin: 900,
    satMargin: 900,
    sunMargin: 900,
  };

  const address0 = {
    position: 0,
    lon: 43.7102,
    lat: 7.262,
    locality: "'Nice'",
    postalCode: "'06000'",
    country: "'France'",
  };
  const address1 = {
    position: 1,
    lon: 43.2965,
    lat: 5.3698,
    locality: "'Marseille'",
    postalCode: "'13000'",
    country: "'France'",
  };
  const createPunctualDriverAds = async (nbToCreate = 10) => {
    const adToCreate = {
      ...baseUuid,
      ...baseUserUuid,
      ...driverAd,
      ...punctualAd,
    };
    for (let i = 0; i < nbToCreate; i++) {
      adToCreate.uuid = getSeed(i, baseUuid.uuid);
      await executeInsertCommand('ad', adToCreate);
      await executeInsertCommand('address', {
        uuid: getSeed(i, baseAdress0Uuid.uuid),
        adUuid: adToCreate.uuid,
        ...address0,
      });
      await executeInsertCommand('address', {
        uuid: getSeed(i, baseAdress1Uuid.uuid),
        adUuid: adToCreate.uuid,
        ...address1,
      });
    }
  };
  const createRecurrentDriverAds = async (nbToCreate = 10) => {
    const adToCreate = {
      ...baseUuid,
      ...baseUserUuid,
      ...driverAd,
      ...punctualAd,
    };
    for (let i = 0; i < nbToCreate; i++) {
      adToCreate.uuid = getSeed(i, baseUuid.uuid);
      await executeInsertCommand('ad', adToCreate);
      await executeInsertCommand('address', {
        uuid: getSeed(i, baseAdress0Uuid.uuid),
        adUuid: adToCreate.uuid,
        ...address0,
      });
      await executeInsertCommand('address', {
        uuid: getSeed(i, baseAdress1Uuid.uuid),
        adUuid: adToCreate.uuid,
        ...address1,
      });
    }
  };

  const createPunctualPassengerAds = async (nbToCreate = 10) => {
    const adToCreate = {
      ...baseUuid,
      ...baseUserUuid,
      ...passengerAd,
      ...punctualAd,
    };
    for (let i = 0; i < nbToCreate; i++) {
      adToCreate.uuid = getSeed(i, baseUuid.uuid);
      await executeInsertCommand('ad', adToCreate);
      await executeInsertCommand('address', {
        uuid: getSeed(i, baseAdress0Uuid.uuid),
        adUuid: adToCreate.uuid,
        ...address0,
      });
      await executeInsertCommand('address', {
        uuid: getSeed(i, baseAdress1Uuid.uuid),
        adUuid: adToCreate.uuid,
        ...address1,
      });
    }
  };

  const createRecurrentPassengerAds = async (nbToCreate = 10) => {
    const adToCreate = {
      ...baseUuid,
      ...baseUserUuid,
      ...passengerAd,
      ...recurrentAd,
    };
    for (let i = 0; i < nbToCreate; i++) {
      adToCreate.uuid = getSeed(i, baseUuid.uuid);
      await executeInsertCommand('ad', adToCreate);
      await executeInsertCommand('address', {
        uuid: getSeed(i, baseAdress0Uuid.uuid),
        adUuid: adToCreate.uuid,
        ...address0,
      });
      await executeInsertCommand('address', {
        uuid: getSeed(i, baseAdress1Uuid.uuid),
        adUuid: adToCreate.uuid,
        ...address1,
      });
    }
  };

  const createPunctualDriverPassengerAds = async (nbToCreate = 10) => {
    const adToCreate = {
      ...baseUuid,
      ...baseUserUuid,
      ...driverAndPassengerAd,
      ...punctualAd,
    };
    for (let i = 0; i < nbToCreate; i++) {
      adToCreate.uuid = getSeed(i, baseUuid.uuid);
      await executeInsertCommand('ad', adToCreate);
      await executeInsertCommand('address', {
        uuid: getSeed(i, baseAdress0Uuid.uuid),
        adUuid: adToCreate.uuid,
        ...address0,
      });
      await executeInsertCommand('address', {
        uuid: getSeed(i, baseAdress1Uuid.uuid),
        adUuid: adToCreate.uuid,
        ...address1,
      });
    }
  };

  const createRecurrentDriverPassengerAds = async (nbToCreate = 10) => {
    const adToCreate = {
      ...baseUuid,
      ...baseUserUuid,
      ...driverAndPassengerAd,
      ...recurrentAd,
    };
    for (let i = 0; i < nbToCreate; i++) {
      adToCreate.uuid = getSeed(i, baseUuid.uuid);
      await executeInsertCommand('ad', adToCreate);
      await executeInsertCommand('address', {
        uuid: getSeed(i, baseAdress0Uuid.uuid),
        adUuid: adToCreate.uuid,
        ...address0,
      });
      await executeInsertCommand('address', {
        uuid: getSeed(i, baseAdress1Uuid.uuid),
        adUuid: adToCreate.uuid,
        ...address1,
      });
    }
  };
  beforeAll(async () => {
    const module = await Test.createTestingModule({
      imports: [DatabaseModule],
      providers: [PrismaService, AdsRepository],
    }).compile();
    prismaService = module.get<PrismaService>(PrismaService);
    adsRepository = module.get<AdsRepository>(AdsRepository);
  });
  afterAll(async () => {
    await prismaService.$disconnect();
  });

  beforeEach(async () => {
    await prismaService.ad.deleteMany();
  });
  describe('findAll', () => {
    it('should return an empty data array', async () => {
      const res = await adsRepository.findAll();
      expect(res).toEqual({
        data: [],
        total: 0,
      });
    });

    describe('drivers', () => {
      it('should return a data array with 8 punctual driver ads', async () => {
        await createPunctualDriverAds(8);
        const ads = await adsRepository.findAll();
        expect(ads.data.length).toBe(8);
        expect(ads.total).toBe(8);
        expect(ads.data[0].driver).toBeTruthy();
        expect(ads.data[0].passenger).toBeFalsy();
      });

      it('should return a data array limited to 10 punctual driver ads', async () => {
        await createPunctualDriverAds(20);
        const ads = await adsRepository.findAll();
        expect(ads.data.length).toBe(10);
        expect(ads.total).toBe(20);
        expect(ads.data[1].driver).toBeTruthy();
        expect(ads.data[1].passenger).toBeFalsy();
      });

      it('should return a data array with 8 recurrent driver ads', async () => {
        await createRecurrentDriverAds(8);
        const ads = await adsRepository.findAll();
        expect(ads.data.length).toBe(8);
        expect(ads.total).toBe(8);
        expect(ads.data[2].driver).toBeTruthy();
        expect(ads.data[2].passenger).toBeFalsy();
      });

      it('should return a data array limited to 10 recurrent driver ads', async () => {
        await createRecurrentDriverAds(20);
        const ads = await adsRepository.findAll();
        expect(ads.data.length).toBe(10);
        expect(ads.total).toBe(20);
        expect(ads.data[3].driver).toBeTruthy();
        expect(ads.data[3].passenger).toBeFalsy();
      });
    });

    describe('passengers', () => {
      it('should return a data array with 7 punctual passenger ads', async () => {
        await createPunctualPassengerAds(7);
        const ads = await adsRepository.findAll();
        expect(ads.data.length).toBe(7);
        expect(ads.total).toBe(7);
        expect(ads.data[0].passenger).toBeTruthy();
        expect(ads.data[0].driver).toBeFalsy();
      });

      it('should return a data array limited to 10 punctual passenger ads', async () => {
        await createPunctualPassengerAds(15);
        const ads = await adsRepository.findAll();
        expect(ads.data.length).toBe(10);
        expect(ads.total).toBe(15);
        expect(ads.data[1].passenger).toBeTruthy();
        expect(ads.data[1].driver).toBeFalsy();
      });

      it('should return a data array with 7 recurrent passenger ads', async () => {
        await createRecurrentPassengerAds(7);
        const ads = await adsRepository.findAll();
        expect(ads.data.length).toBe(7);
        expect(ads.total).toBe(7);
        expect(ads.data[2].passenger).toBeTruthy();
        expect(ads.data[2].driver).toBeFalsy();
      });

      it('should return a data array limited to 10 recurrent passenger ads', async () => {
        await createRecurrentPassengerAds(15);
        const ads = await adsRepository.findAll();
        expect(ads.data.length).toBe(10);
        expect(ads.total).toBe(15);
        expect(ads.data[3].passenger).toBeTruthy();
        expect(ads.data[3].driver).toBeFalsy();
      });
    });

    describe('drivers and passengers', () => {
      it('should return a data array with 6 punctual driver and passenger ads', async () => {
        await createPunctualDriverPassengerAds(6);
        const ads = await adsRepository.findAll();
        expect(ads.data.length).toBe(6);
        expect(ads.total).toBe(6);
        expect(ads.data[0].passenger).toBeTruthy();
        expect(ads.data[0].driver).toBeTruthy();
      });

      it('should return a data array limited to 10 punctual driver and passenger ads', async () => {
        await createPunctualDriverPassengerAds(16);
        const ads = await adsRepository.findAll();
        expect(ads.data.length).toBe(10);
        expect(ads.total).toBe(16);
        expect(ads.data[1].passenger).toBeTruthy();
        expect(ads.data[1].driver).toBeTruthy();
      });

      it('should return a data array with 6 recurrent driver and passenger ads', async () => {
        await createRecurrentDriverPassengerAds(6);
        const ads = await adsRepository.findAll();
        expect(ads.data.length).toBe(6);
        expect(ads.total).toBe(6);
        expect(ads.data[2].passenger).toBeTruthy();
        expect(ads.data[2].driver).toBeTruthy();
      });

      it('should return a data array limited to 10 recurrent driver and passenger ads', async () => {
        await createRecurrentDriverPassengerAds(16);
        const ads = await adsRepository.findAll();
        expect(ads.data.length).toBe(10);
        expect(ads.total).toBe(16);
        expect(ads.data[3].passenger).toBeTruthy();
        expect(ads.data[3].driver).toBeTruthy();
      });
    });
  });
  describe('findOneByUuid', () => {
    it('should return an ad', async () => {
      await createPunctualDriverAds(1);
      const ad = await adsRepository.findOneByUuid(baseUuid.uuid);

      expect(ad.uuid).toBe(baseUuid.uuid);
    });

    it('should return null', async () => {
      const ad = await adsRepository.findOneByUuid(
        '544572be-11fb-4244-8235-587221fc9104',
      );
      expect(ad).toBeNull();
    });
  });
  describe('create', () => {
    it('should create an  punctual ad', async () => {
      const beforeCount = await prismaService.ad.count();
      const adToCreate: AdDTO = new AdDTO();

      adToCreate.uuid = 'be459a29-7a41-4c0b-b371-abe90bfb6f00';
      adToCreate.userUuid = '4e52b54d-a729-4dbd-9283-f84a11bb2200';
      adToCreate.driver = true;
      adToCreate.passenger = false;
      adToCreate.frequency = Frequency.PUNCTUAL;
      adToCreate.fromDate = new Date('05-22-2023 09:36');
      adToCreate.toDate = new Date('05-22-2023 09:36');
      adToCreate.monTime = '09:36';
      adToCreate.monMargin = 900;
      adToCreate.tueMargin = 900;
      adToCreate.wedMargin = 900;
      adToCreate.thuMargin = 900;
      adToCreate.friMargin = 900;
      adToCreate.satMargin = 900;
      adToCreate.sunMargin = 900;
      adToCreate.seatsProposed = 3;
      adToCreate.seatsRequested = 0;
      adToCreate.strict = false;
      adToCreate.waypoints = {
        create: [address0 as Waypoint, address1 as Waypoint],
      };
      const ad = await adsRepository.create(adToCreate);

      const afterCount = await prismaService.ad.count();

      expect(afterCount - beforeCount).toBe(1);
      expect(ad.uuid).toBe('be459a29-7a41-4c0b-b371-abe90bfb6f00');
    });
    it('should create an  recurrent ad', async () => {
      const beforeCount = await prismaService.ad.count();
      const adToCreate: AdDTO = new AdDTO();

      adToCreate.uuid = '137a26fa-4b38-48ba-aecf-1a75f6b20f3d';
      adToCreate.userUuid = '4e52b54d-a729-4dbd-9283-f84a11bb2200';
      adToCreate.driver = true;
      adToCreate.passenger = false;
      adToCreate.frequency = Frequency.RECURRENT;
      adToCreate.fromDate = new Date('01-15-2023 ');
      adToCreate.toDate = new Date('10-31-2023');
      adToCreate.monTime = '07:30';
      adToCreate.friTime = '07:45';
      adToCreate.thuTime = '08:00';
      adToCreate.monMargin = 900;
      adToCreate.tueMargin = 900;
      adToCreate.wedMargin = 900;
      adToCreate.thuMargin = 900;
      adToCreate.friMargin = 900;
      adToCreate.satMargin = 900;
      adToCreate.sunMargin = 900;
      adToCreate.seatsProposed = 2;
      adToCreate.seatsRequested = 0;
      adToCreate.strict = false;
      adToCreate.waypoints = {
        create: [address0 as Waypoint, address1 as Waypoint],
      };
      const ad = await adsRepository.create(adToCreate);

      const afterCount = await prismaService.ad.count();

      expect(afterCount - beforeCount).toBe(1);
      expect(ad.uuid).toBe('137a26fa-4b38-48ba-aecf-1a75f6b20f3d');
    });
  });
});
